from dataclasses import dataclass
import requests
from pathlib import Path
from typing import Optional, Iterable
import re
from urllib.parse import parse_qs, urlencode
import os 
from .url import URL

SelectType = Iterable[str] | str
FilterType = str

def _parse_creds(host: str | None = None, username: str | None = None, password: str | None = None, ssl=None):
    url = os.getenv("DENODO_URL")
    if url:
        parsed_url = URL.from_url(url)
        host = host or parsed_url.host
        ssl = ssl if ssl is not None else parsed_url.ssl
        if parsed_url.credentials:
            username = username or parsed_url.credentials.username
            password = password or parsed_url.credentials.password
    host = host or os.getenv("DENODO_HOST")
    username = username or os.getenv("DENODO_USERNAME")
    password = password or os.getenv("DENODO_PASSWORD")
    return host, username, password, ssl

class Client:
    def __init__(self, host: str | None = None, username: str | None = None, password: str | None = None, verify:Optional[str | Path | bool]=None, ssl=None):
        host, username, password, ssl = _parse_creds(host, username, password)
        if isinstance(verify, str):
            verify = Path(verify)
        # ---
        # Confirm that the certificat is present
        if isinstance(verify, Path):
            if not verify.is_file():
                # logging.warning("\n".join(str(p) for p in Path().iterdir()))
                raise Exception("Certificate not found")
            verify = str(verify)
        
        session = requests.Session()

        session.auth = (username, password)
        if verify is not None:
            session.verify = verify
        
        session.headers.update({
            "Content-Type":"application/json",
            "Accept":"application/json"
        })
        self._session: requests.Session = session
        self._host: str = host
    
    @property
    def session(self) -> requests.Session:
        return self._session
    
    # @property
    # def host(self) -> str:
    #     return self._host

    @property
    def url(self) -> str:
        return "https://denodo-prod02.hq.k.grp:9443"
    
    # @staticmethod
    # def from_url(self, url: str) -> "Client":
    #     parsed = URL.from_url(url)
    #     return

    def database(self, name: str) -> "Database":
        return Database(self, name)


class Database:
    def __init__(self, client: Client, name: str):
        self._client: Client = client
        self._name: str = name

    @property
    def client(self) -> Client:
        return self._client
    
    @property
    def name(self) -> str:
        return self._name

    @property
    def restful_url(self) -> str:
        # NOTE: This is the new RESTFul API
        # it is simpler than the previous REST API
        return f"{self._client.url}/denodo-restfulws/{self.name}"
    
    url = restful_url
    
    def view(self, name: str) -> "View":
        return View(self, name)

class View:
    def __init__(self, database: Database, name: str):
        self._database: Database = database
        self._name: str = name

    @property
    def database(self) -> Database:
        return self._database

    @property
    def client(self) -> Client:
        return self._database.client
    
    @property
    def name(self) -> str:
        return self._name

    @property
    def restful_url(self) -> str:
        # NOTE: This is the new RESTFul API
        # it is simpler than the previous REST API
        return f"{self.database.restful_url}/views/{self.name}"
    
    url = restful_url
    
    def _query_denodo(self, select:Optional[str]=None, filter: Optional[str] =None) -> dict:
        params = {
            "$format": "json", 
        }
        if select is not None:
            params["$select"] = select 
        if filter is not None:
            params["$filter"] = filter 
        
        response = self.client.session.get(
            url=self.url,
            params=params,
        )

        response.raise_for_status()
        # if response.status_code == 401:
        #     raise Exception(f"From Denodo (HTTP {response.status_code}): You are not authorized to access this resource.")
        # if response.status_code != 200:
        #     raise Exception(f"From Denodo (HTTP {response.status_code}): {response.content.decode()}")
        
        data = response.json()
        errors = data.get("__errors__")
        if not errors:
            return data 
        raise Exception(f"From Denodo: {errors}")


    def get(self, select: Optional[SelectType] = None, filter: Optional[FilterType] = None) -> list[dict]:
        if select is not None and not isinstance(select, str):
            select = ",".join(f.strip() for f in select)
        data = self._query_denodo(
            select=select,
            filter=filter,
        )
        elements = data.get("elements") or []
        return elements

